require "defines"

MAX_CONFIG_SIZE = 10
MAX_STORAGE_SIZE = 12

function glob_init()

    glob["entity-recipes"] = glob["entity-recipes"] or {}
    glob["config"] = glob["config"] or {}
    glob["config-tmp"] = glob["config-tmp"] or {}
    glob["storage"] = glob["storage"] or {}

end

function get_type(entity)

    if game.entityprototypes[entity] then
        return game.entityprototypes[entity].type
    end
    return ""

end

function count_keys(hashmap)

    local result = 0

    for _, __ in pairs(hashmap) do
        result = result + 1
    end

    return result

end

-- Hacky function to determine whether transport belt to ground is input
-- or output. It is based on a fact that position of input on the map is
-- slightly different than position of output.

function get_input_output(entity)

    local x = entity.position.x
    local y = entity.position.y
    local dir = entity.direction
    local i = 0
    local j = 0

    if dir == defines.direction.north then
        i, j = math.modf(y)
        if j == -0.75 or j == 0.25 then return "input" end
    elseif dir == defines.direction.east then
        i, j = math.modf(x)
        if j == -0.25 or j == 0.75 then return "input" end
    elseif dir == defines.direction.west then
        i, j = math.modf(y)
        if j == -0.25 or j == 0.75 then return "input" end
    else
        i, j = math.modf(x)
        if j == -0.75 or j == 0.25 then return "input" end
    end

    return "output"

end

function get_config_item(player, index, type)

    if not glob["config-tmp"][player.name]
            or index > #glob["config-tmp"][player.name]
            or glob["config-tmp"][player.name][index][type] == "" then

        return game.localise("replacer-item-not-set")

    end

    return game.getlocaliseditemname(glob["config-tmp"][player.name][index][type])

end

function is_item_valid(entity)

    return (game.entityprototypes[entity]
        and game.entityprototypes[entity].hasflag("placeable-neutral")
        and game.entityprototypes[entity].hasflag("player-creation"))

end

function gui_init(player, after_research)

    if not player.gui.top["replacer-config-button"]
            and (player.force.technologies["automated-construction"].researched or after_research) then

        player.gui.top.add{
            type = "button",
            name = "replacer-config-button",
            caption = game.localise("replacer-config-button-caption")
        }

    end

end

function gui_open_frame(player)

    local frame = player.gui.left["replacer-config-frame"]
    local storage_frame = player.gui.left["replacer-storage-frame"]

    if frame then
        frame.destroy()
        if storage_frame then
            storage_frame.destroy()
        end
        glob["config-tmp"][player.name] = nil
        return
    end

    -- If player config does not exist, we need to create it.

    glob["config"][player.name] = glob["config"][player.name] or {}

    -- Temporary config lives as long as the frame is open, so it has to be created
    -- every time the frame is opened.

    glob["config-tmp"][player.name] = {}

    -- We need to copy all items from normal config to temporary config.

    local i = 0

    for i = 1, MAX_CONFIG_SIZE do

        if i > #glob["config"][player.name] then
            glob["config-tmp"][player.name][i] = { from = "", to = "" }
        else
            glob["config-tmp"][player.name][i] = {
                from = glob["config"][player.name][i].from, 
                to = glob["config"][player.name][i].to
            }
        end
        
    end

    -- Now we can build the GUI.

    frame = player.gui.left.add{
        type = "frame",
        caption = game.localise("replacer-config-frame-title"),
        name = "replacer-config-frame",
        direction = "vertical"
    }

    local error_label = frame.add{ 
        type = "label",
        caption = "---",
        name = "replacer-error-label"
    }

    error_label.style.minimalwidth = 200

    local ruleset_grid = frame.add{
        type = "table",
        colspan = 3,
        name = "replacer-ruleset-grid"
    }

    ruleset_grid.add{
        type = "label",
        name = "replacer-grid-header-1",
        caption = game.localise("replacer-config-header-1")
    }
    ruleset_grid.add{
        type = "label",
        name = "replacer-grid-header-2",
        caption = game.localise("replacer-config-header-2")
    }
    ruleset_grid.add{
        type = "label",
        name = "replacer-grid-header-3",
        caption = ""
    }

    for i = 1, MAX_CONFIG_SIZE do
        ruleset_grid.add{ 
            type = "button",
            name = "replacer-from-" .. i,
            style = "replacer-small-button",
            caption = get_config_item(player, i, "from")
        }
        ruleset_grid.add{
            type = "button",
            name = "replacer-to-" .. i,
            style = "replacer-small-button",
            caption = get_config_item(player, i, "to")
        }
        ruleset_grid.add{
            type = "button",
            name = "replacer-clear-" .. i,
            style = "replacer-small-button",
            caption = game.localise("replacer-config-button-clear")
        }
    end

    local button_grid = frame.add{
        type = "table",
        colspan = 2,
        name = "replacer-button-grid"
    }

    button_grid.add{
        type = "button",
        name = "replacer-apply",
        caption = game.localise("replacer-config-button-apply")
    }
    button_grid.add{
        type = "button",
        name = "replacer-clear-all",
        caption = game.localise("replacer-config-button-clear-all")
    }

    storage_frame = player.gui.left.add{
        type = "frame",
        name = "replacer-storage-frame",
        caption = game.localise("replacer-storage-frame-title"),
        direction = "vertical"
    }

    local storage_frame_error_label = storage_frame.add{
        type = "label",
        name = "replacer-storage-error-label",
        caption = "---"
    }

    storage_frame_error_label.style.minimalwidth = 200

    local storage_frame_buttons = storage_frame.add{
        type = "table",
        colspan = 3,
        name = "replacer-storage-buttons"
    }

    storage_frame_buttons.add{
        type = "label",
        caption = game.localise("replacer-storage-name-label"),
        name = "replacer-storage-name-label"
    }

    storage_frame_buttons.add{
        type = "textfield",
        text = "",
        name = "replacer-storage-name"
    }

    storage_frame_buttons.add{
        type = "button",
        caption = game.localise("replacer-storage-store"),
        name = "replacer-storage-store",
        style = "replacer-small-button"
    }

    local storage_grid = storage_frame.add{
        type = "table",
        colspan = 3,
        name = "replacer-storage-grid"
    }

    if glob["storage"][player.name] then

        i = 1
        for key, _ in pairs(glob["storage"][player.name]) do
            storage_grid.add{
                type = "label",
                caption = key .. "        ",
                name = "replacer-storage-entry-" .. i
            }

            storage_grid.add{
                type = "button",
                caption = game.localise("replacer-storage-restore"),
                name = "replacer-restore-" .. i,
                style = "replacer-small-button"
            }
            storage_grid.add{
                type = "button",
                caption = game.localise("replacer-storage-remove"),
                name = "replacer-remove-" .. i,
                style = "replacer-small-button"
            }
            i = i + 1
        end

    end

end

function gui_save_changes(player)

    -- Saving changes consists in:
    --   1. copying config-tmp to config
    --   2. removing config-tmp
    --   3. closing the frame

    if glob["config-tmp"][player.name] then

        local i = 0
        glob["config"][player.name] = {}

        for i = 1, #glob["config-tmp"][player.name] do

            -- Rule can be saved only if both "from" and "to" fields are set.

            if glob["config-tmp"][player.name][i].from == ""
                    or glob["config-tmp"][player.name][i].to == "" then

                glob["config"][player.name][i] = { from = "", to = "" }

            else
                glob["config"][player.name][i] = {
                    from = glob["config-tmp"][player.name][i].from,
                    to = glob["config-tmp"][player.name][i].to
                }
            end
            
        end

        glob["config-tmp"][player.name] = nil

    end

    local frame = player.gui.left["replacer-config-frame"]
    local storage_frame = player.gui.left["replacer-storage-frame"]

    if frame then
        frame.destroy()
        if storage_frame then
            storage_frame.destroy()
        end
    end

end

function gui_clear_all(player)

    local i = 0
    local frame = player.gui.left["replacer-config-frame"]

    if not frame then return end

    local ruleset_grid = frame["replacer-ruleset-grid"]

    for i = 1, MAX_CONFIG_SIZE do

        glob["config-tmp"][player.name][i] = { from = "", to = "" }
        ruleset_grid["replacer-from-" .. i].caption = game.localise("replacer-item-not-set")
        ruleset_grid["replacer-to-" .. i].caption = game.localise("replacer-item-not-set")
        
    end

end

function gui_display_message(frame, storage, message)

    local label_name = "replacer-"
    if storage then label_name = label_name .. "storage-" end
    label_name = label_name .. "error-label"

    local error_label = frame[label_name]
    if not error_label then return end

    if message ~= "---" then
        message = game.localise(message)
    end

    error_label.caption = message

end

function gui_set_rule(player, type, index)

    local frame = player.gui.left["replacer-config-frame"]
    if not frame or not glob["config-tmp"][player.name] then return end

    local stack = player.cursorstack

    if not stack or not stack.name then 
        gui_display_message(frame, false, "replacer-item-empty")
        return
    end

    if not is_item_valid(stack.name) then
        gui_display_message(frame, false, "replacer-item-not-valid")
        return
    end

    local opposite = "from"
    local i = 0

    if type == "from" then

        opposite = "to"

        for i = 1, #glob["config-tmp"][player.name] do
            if index ~= i and glob["config-tmp"][player.name][i].from == stack.name then
                gui_display_message(frame, false, "replacer-item-already-set")
                return
            end
        end

    end

    local related = glob["config-tmp"][player.name][index][opposite]

    if related ~= "" then

        if related == stack.name then
            gui_display_message(frame, false, "replacer-item-is-same")
            return
        end

        if get_type(stack.name) ~= get_type(related) then
            gui_display_message(frame, false, "replacer-item-not-same-type")
            return
        end

    end

    glob["config-tmp"][player.name][index][type] = stack.name

    local ruleset_grid = frame["replacer-ruleset-grid"]
    ruleset_grid["replacer-" .. type .. "-" .. index].caption = game.getlocaliseditemname(stack.name)

end

function gui_clear_rule(player, index)

    local frame = player.gui.left["replacer-config-frame"]
    if not frame or not glob["config-tmp"][player.name] then return end

    gui_display_message(frame, false, "---")

    local ruleset_grid = frame["replacer-ruleset-grid"]

    glob["config-tmp"][player.name][index] = { from = "", to = "" }
    ruleset_grid["replacer-from-" .. index].caption = game.localise("replacer-item-not-set")
    ruleset_grid["replacer-to-" .. index].caption = game.localise("replacer-item-not-set")

end

function gui_store(player)

    glob["storage"][player.name] = glob["storage"][player.name] or {}

    local storage_frame = player.gui.left["replacer-storage-frame"]
    if not storage_frame then return end

    local textfield = storage_frame["replacer-storage-buttons"]["replacer-storage-name"]
    local name = textfield.text
    name = string.match(name, "^%s*(.-)%s*$")

    if not name or name == "" then
        gui_display_message(storage_frame, true, "replacer-storage-name-not-set")
        return
    end

    if glob["storage"][player.name][name] then
        gui_display_message(storage_frame, true, "replacer-storage-name-in-use")
        return
    end

    glob["storage"][player.name][name] = {}
    local i = 0

    for i = 1, #glob["config-tmp"][player.name] do
        glob["storage"][player.name][name][i] = {
            from = glob["config-tmp"][player.name][i].from,
            to = glob["config-tmp"][player.name][i].to
        }
    end

    local storage_grid = storage_frame["replacer-storage-grid"]
    local index = count_keys(glob["storage"][player.name]) + 1

    if index > MAX_STORAGE_SIZE + 1 then
        gui_display_message(storage_frame, true, "replacer-storage-too-long")
        return
    end

    storage_grid.add{
        type = "label",
        caption = name .. "        ",
        name = "replacer-storage-entry-" .. index
    }

    storage_grid.add{
        type = "button",
        caption = game.localise("replacer-storage-restore"),
        name = "replacer-restore-" .. index,
        style = "replacer-small-button"
    }

    storage_grid.add{
        type = "button",
        caption = game.localise("replacer-storage-remove"),
        name = "replacer-remove-" .. index,
        style = "replacer-small-button"
    }

    gui_display_message(storage_frame, true, "---")
    textfield.text = ""

end

function gui_restore(player, index)

    local frame = player.gui.left["replacer-config-frame"]
    local storage_frame = player.gui.left["replacer-storage-frame"]
    if not frame or not storage_frame then return end

    local storage_grid = storage_frame["replacer-storage-grid"]
    local storage_entry = storage_grid["replacer-storage-entry-" .. index]
    if not storage_entry then return end

    local name = string.match(storage_entry.caption, "^%s*(.-)%s*$")
    if not glob["storage"][player.name] or not glob["storage"][player.name][name] then return end

    glob["config-tmp"][player.name] = {}

    local i = 0
    local ruleset_grid = frame["replacer-ruleset-grid"]

    for i = 1, MAX_CONFIG_SIZE do
        if i > #glob["storage"][player.name][name] then
            glob["config-tmp"][player.name][i] = { from = "", to = "" }
        else
            glob["config-tmp"][player.name][i] = {
                from = glob["storage"][player.name][name][i].from,
                to = glob["storage"][player.name][name][i].to
            }
        end
        ruleset_grid["replacer-from-" .. i].caption = get_config_item(player, i, "from")
        ruleset_grid["replacer-to-" .. i].caption = get_config_item(player, i, "to")
    end

    gui_display_message(storage_frame, true, "---")

end

function gui_remove(player, index)

    if not glob["storage"][player.name] then return end

    local storage_frame = player.gui.left["replacer-storage-frame"]
    if not storage_frame then return end

    local storage_grid = storage_frame["replacer-storage-grid"]
    local label = storage_grid["replacer-storage-entry-" .. index]
    local btn1 = storage_grid["replacer-restore-" .. index]
    local btn2 = storage_grid["replacer-remove-" .. index]

    if not label or not btn1 or not btn2 then return end

    local name = string.match(label.caption, "^%s*(.-)%s*$")

    label.destroy()
    btn1.destroy()
    btn2.destroy()

    glob["storage"][player.name][name] = nil

    gui_display_message(storage_frame, true, "---")

end

game.onevent(defines.events.onmarkedfordeconstruction, function(event)

    local entity = event.entity
    local deconstruction = false
    local upgrade = false
    local player = nil
    local i = 0

    -- Determine which player used upgrade planner.
    -- If more than one player has upgrade planner in their hand or one
    -- player has a upgrade planner and other has deconstruction planner,
    -- we can't determine it, so we have to discard deconstruction order.

    for i = 1, #game.players do

        if game.players[i].cursorstack then
            if game.players[i].cursorstack.name == "upgrade-planner" then
                if upgrade or deconstruction then
                    entity.canceldeconstruction(entity.force)
                    return
                end
                player = game.players[i]
                upgrade = true
            elseif game.players[i].cursorstack.name == "deconstruction-planner" then
                if upgrade then
                    entity.canceldeconstruction(entity.force)
                    return
                end
                deconstruction = true
            end
        end

    end

    if not player then return end

    -- Get player config.

    if not glob["config"][player.name] then

        -- Config for this player does not exist yet, so we have nothing to do.
        -- We can create it now for later usage.

        glob["config"][player.name] = {}
        entity.canceldeconstruction(entity.force)
        return
        
    end

    local config = glob["config"][player.name]

    -- Check if entity is valid and stored in config as a source.

    local index = 0

    for i = 1, #config do

        if config[i].from == entity.name then
            index = i
            break
        end

    end

    if index == 0 then
        entity.canceldeconstruction(entity.force)
        return
    end

    local type = get_type(entity.name) 
    if type == "" then
        entity.canceldeconstruction(entity.force)
        return
    end

    local new_entity = {
        name = "ghost",
        innername = config[index].to,
        position = entity.position,
        direction = entity.direction,
        force = entity.force
    }

    -- If entity is an assembling machine, we need to preserve its recipe.

    if type == "assembling-machine" then
        glob["entity-recipes"][entity.position.x .. ":" .. entity.position.y] = entity.recipe
    end

    -- If entity is a transport belt to ground, we need to preserve it type (input or output)

    if type == "transport-belt-to-ground" then
        new_entity.type = get_input_output(entity)
    end

    game.createentity(new_entity)

end)

game.onevent(defines.events.onrobotbuiltentity, function(event)

    local entity = event.createdentity

    -- Restore recipe of any assembling machine placed by the robot.

    if get_type(entity.name) == "assembling-machine" then

        local tag = entity.position.x .. ":" .. entity.position.y
        if glob["entity-recipes"][tag] then
            entity.recipe = glob["entity-recipes"][tag]
            glob["entity-recipes"][tag] = nil
        end

    end

end)

game.onevent(defines.events.onguiclick, function(event) 

    local element = event.element
    local player = game.getplayer(event.playerindex)

    if element.name == "replacer-config-button" then
        gui_open_frame(player)
    elseif element.name == "replacer-apply" then
        gui_save_changes(player)
    elseif element.name == "replacer-clear-all" then
        gui_clear_all(player)
    elseif element.name  == "replacer-storage-store" then
        gui_store(player)
    else

        local type, index = string.match(element.name, "replacer%-(%a+)%-(%d+)")
        if type and index then
            if type == "from" or type == "to" then
                gui_set_rule(player, type, tonumber(index))
            elseif type == "restore" then
                gui_restore(player, tonumber(index))
            elseif type == "remove" then
                gui_remove(player, tonumber(index))
            elseif type == "clear" then
                gui_clear_rule(player, tonumber(index))
            end
        end

    end

end)

game.onevent(defines.events.onresearchfinished, function(event)

    if event.research == 'automated-construction' then
        for _, player in pairs(game.players) do
            gui_init(player, true)
        end
    end

end)

game.oninit(function()

    glob_init()

end)

game.onload(function() 

    glob_init()

    for _, player in pairs(game.players) do
        gui_init(player, false)
    end

end)

